package de.unisaarland.cs.st.tautoko;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import de.unisaarland.cs.st.adabu.trans.model.MethodInvocation;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocationSet;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectModel;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectState;
import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.adabu.util.datastructures.Graph;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphEdge;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphNode;

public class ModelToTypeStateConverter {

	public static final String STATE_START = "start";

	public static final String STATE_EXCEPTION = "ex";

	public static Graph<String, List<MethodIdentifier>> convertModel(TransitiveObjectModel model) {
		Map<TransitiveObjectState, String> stateToIdentifierMapping = new HashMap<TransitiveObjectState, String>();
		stateToIdentifierMapping.put(model.getStartingState(), STATE_START);
		int idCounter = 0;
		for (GraphNode<TransitiveObjectState, MethodInvocationSet> node: model.getModel().getNodes()) {
			TransitiveObjectState state = node.getData();
			if (!stateToIdentifierMapping.containsKey(state)) {
				stateToIdentifierMapping.put(state, "node_" + idCounter);
				idCounter++;
			}
		}
		Graph<String, List<MethodIdentifier>> result = new Graph<String, List<MethodIdentifier>>();
		for (String nodeId : stateToIdentifierMapping.values()) {
			result.createNode(nodeId);
		}
		GraphNode<String, List<MethodIdentifier>> exceptionNode = result.createNode(STATE_EXCEPTION);
		for (GraphEdge<TransitiveObjectState, MethodInvocationSet> edge : model.getModel().getEdges()) {
			MethodInvocationSet methodInvocations = edge.getData();
			String startNodeId = stateToIdentifierMapping.get(edge.getStartNode().getData());
			GraphNode<String, List<MethodIdentifier>> startNode = result.getNodeForData(startNodeId);
			String endNodeId = stateToIdentifierMapping.get(edge.getEndNode().getData());
			GraphNode<String, List<MethodIdentifier>> endNode = result.getNodeForData(endNodeId);
			List<MethodIdentifier> invokedMethods = methodInvocations.getInvokedMethods();
			List<MethodIdentifier> regularMethodInvocations = new Vector<MethodIdentifier>(invokedMethods.size());
			for (MethodIdentifier methodIdentifier : invokedMethods) {
				if (hasRegularInvocation(methodIdentifier, methodInvocations)) {
					regularMethodInvocations.add(methodIdentifier);
				}
			}
			if (regularMethodInvocations.size() > 0) {
				result.createEdge(regularMethodInvocations, startNode, endNode);
			}
			List<MethodIdentifier> exceptionalMethodInvocations = new Vector<MethodIdentifier>(invokedMethods.size());
			for (MethodIdentifier methodIdentifier : invokedMethods) {
				if (hasExceptionalInvocation(methodIdentifier, methodInvocations)) {
					exceptionalMethodInvocations.add(methodIdentifier);
				}
			}
			if (exceptionalMethodInvocations.size() > 0) {
				if (!result.containsEdge(startNode, exceptionNode)) {
					result.createEdge(exceptionalMethodInvocations, startNode, exceptionNode);
				} else {
					List<MethodIdentifier> existingEdgeData = startNode.getOutgoingEdgeToNode(exceptionNode).getData();
					for (MethodIdentifier methodIdentifier : exceptionalMethodInvocations) {
						if (!existingEdgeData.contains(methodIdentifier)) {
							existingEdgeData.add(methodIdentifier);
						}
					}
				}
			}	
		}
		return result;
	}

	public static boolean hasRegularInvocation(MethodIdentifier methodIdentifier, MethodInvocationSet methodInvocations) {
		Set<MethodInvocation> currentMethodInvocations = methodInvocations.getAllInvocationsOfMethod(methodIdentifier);
		for (MethodInvocation methodInvocation : currentMethodInvocations) {
			if (!methodInvocation.exceptionRaised()) {
				return true;
			}
		}
		return false;
	}
	
	public static boolean hasExceptionalInvocation(MethodIdentifier methodIdentifier, MethodInvocationSet methodInvocations) {
		Set<MethodInvocation> currentMethodInvocations = methodInvocations.getAllInvocationsOfMethod(methodIdentifier);
		for (MethodInvocation methodInvocation : currentMethodInvocations) {
			if (methodInvocation.exceptionRaised()) {
				return true;
			}
		}
		return false;
	}
}
