package de.unisaarland.cs.st.tautoko;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.adabu.util.datastructures.Graph;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphEdge;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphNode;


public class TypeStateGraphMLSerializer {

	public static String getGraphMLRepresentation(Map<String, Graph<String, List<MethodIdentifier>>> typeToTypeStateMap, Map<String, Collection<String>> interfaces, Map<String, Collection<String>> superClasses) {
		String encoding = System.getProperty("file.encoding");
		StringBuffer result = new StringBuffer("<?xml version=\"1.0\" encoding=\"" + encoding + "\"?>\n");
		result.append("<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " + 
      "xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd\">\n");
		for (String className : typeToTypeStateMap.keySet()) {
			Graph<String, List<MethodIdentifier>> typeState = typeToTypeStateMap.get(className);
			String graphMLRepresentation = getGraphMLRepresentation(className, typeState, interfaces.get(className), superClasses.get(className));
			result.append(graphMLRepresentation);
		}
		result.append("</graphml>\n");
		return result.toString();
	}
	
	public static String getGraphMLRepresentation(String className, Graph<String, List<MethodIdentifier>> typeState, Collection<String> interfaces, Collection<String> superClasses) {
		String typeString = className;
		if (interfaces != null) {
			for (String interfaceName : interfaces) {
				typeString = typeString + " " + interfaceName;
			}
		}
		if (superClasses != null) {
			for (String superClassName : superClasses) {
				typeString = typeString + " " + superClassName;
			}
		}
		StringBuffer result = new StringBuffer("<graph id=\"" + typeString + "\" edgedefault=\"directed\">\n");
		result.append("\t<key id=\"key1\" for=\"edge\" attr.name=\"methodidentifiers\" attr.type=\"string\"/>\n");
		for (GraphNode<String, List<MethodIdentifier>> node: typeState.getNodes()) {
			result.append("\t<node id=\"" + node.getData() + "\"/>\n");
		}
		int edgeCounter = 0;
		for (GraphEdge<String, List<MethodIdentifier>> edge : typeState.getEdges()) {
			result.append("\t<edge id=\"e" + edgeCounter + "\" source=\"" + edge.getStartNode().getData() + "\" target=\"" + edge.getEndNode().getData() + "\">\n");
			result.append("\t\t<data key=\"key1\">");
			result.append(toString(edge.getData()));
			result.append("</data>\n");
			result.append("\t</edge>\n");
			edgeCounter++;
		}
		result.append("</graph>\n");
		return result.toString();
	}
	
	public static String toString(List<MethodIdentifier> identifiers) {
		StringBuffer result = new StringBuffer();
		for (int i = 0; i < identifiers.size(); i++) {
			result.append(identifiers.get(i).toString().replace("<init>", "[init]"));
			if (i != identifiers.size() - 1) {
				result.append("|");
			}
		}
		return result.toString();
	}
}
