package de.unisaarland.cs.st.tautoko;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import de.unisaarland.cs.st.adabu.core.state.Value;
import de.unisaarland.cs.st.adabu.trans.abstraction.OldAdabuStateAbstracter;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocation;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocationSet;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectModel;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectState;
import de.unisaarland.cs.st.adabu.util.abstraction.SimpleStateAbstracter;
import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.adabu.util.datastructures.Graph;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphEdge;
import de.unisaarland.cs.st.adabu.util.datastructures.Graph.EdgeDataMerger;


public class Util {


	@SuppressWarnings("unchecked")
	public static TransitiveObjectModel mergeModels(Set<TransitiveObjectModel> models) {
		if (models.size() == 0) {
			return null;
		}
		if (models.size() == 1) {
			return models.iterator().next();
		} else {
			Iterator<TransitiveObjectModel> modelIterator = models.iterator();
			TransitiveObjectModel firstModel = modelIterator.next();
			Graph<TransitiveObjectState, MethodInvocationSet>	currentGraph = firstModel.getModel();
			while (modelIterator.hasNext()) {
				TransitiveObjectModel nextModel = modelIterator.next();
				Graph<TransitiveObjectState, MethodInvocationSet>	graphToMerge = nextModel.getModel();
				Graph<TransitiveObjectState, MethodInvocationSet>	mergedGraph = currentGraph.mergeGraph(graphToMerge, new EdgeDataMerger() {
					public Object mergeEdgeData(Object firstData, Object secondData) {
						MethodInvocationSet firstSet = (MethodInvocationSet) firstData;
						MethodInvocationSet secondSet = (MethodInvocationSet) secondData;
						MethodInvocationSet mergedSet = new MethodInvocationSet();
						mergedSet.addAll(firstSet);
						mergedSet.addAll(secondSet);
						return mergedSet;
					}
				});
				currentGraph = mergedGraph;
			}
			TransitiveObjectModel result = new TransitiveObjectModel(firstModel.getClassName(), 0, currentGraph);
			result.setStartingState(firstModel.getStartingState());
			result.setCurrentState(firstModel.getCurrentState());
			return result;
		}
	}
	public static void removeSpuriousInvocations(Map<String, TransitiveObjectModel> typeToMergedModelMap) {
		for (String type : typeToMergedModelMap.keySet()) {
			TransitiveObjectModel model = typeToMergedModelMap.get(type);
			for (GraphEdge<TransitiveObjectState, MethodInvocationSet> edge: model.getModel().getEdges()) {
				MethodInvocationSet invocationSet = edge.getData();
				Map<MethodIdentifier, List<MethodInvocation>> methodToInvocationMap = getMethodToInvocationMap(invocationSet);
				for (MethodIdentifier methodIdentifier: methodToInvocationMap.keySet()) {
					List<MethodInvocation> currentMethodInvocations = methodToInvocationMap.get(methodIdentifier);
					if (currentMethodInvocations.size() > 1) {
						Map<MethodInvocationContext, MethodInvocation> contextToInvocationMap = new HashMap<MethodInvocationContext, MethodInvocation>();
						for (MethodInvocation methodInvocation : currentMethodInvocations) {
							Map<Integer, Value> abstractParameters = getAbstractParameterValues(methodInvocation.getParameters());
							MethodInvocationContext context = new MethodInvocationContext(abstractParameters, methodInvocation.getExceptionClass());
							if (!contextToInvocationMap.containsKey(context)) {
								contextToInvocationMap.put(context, methodInvocation);
							} else {
								invocationSet.removeInvocation(methodInvocation);
							}
						}
					}
				}
			}
		}
	}	
	
	public static Map<Integer, Value> getAbstractParameterValues(Map<Integer, Value> parameters) {
		Map<Integer, Value> result = new HashMap<Integer, Value>();
		for (Integer parameterId : parameters.keySet()) {
			result.put(parameterId, SimpleStateAbstracter.getAbstractValue(parameters.get(parameterId)));
		}
		return result;
	}
	


	public static Map<MethodIdentifier, List<MethodInvocation>> getMethodToInvocationMap(
																																												MethodInvocationSet invocationSet) {
		Map<MethodIdentifier, List<MethodInvocation>> methodToInvocationMap = new HashMap<MethodIdentifier, List<MethodInvocation>>();
		for (MethodInvocation methodInvocation : invocationSet) {
			List<MethodInvocation> currentMethodInvocations = methodToInvocationMap.get(methodInvocation.getIdentifier());
			if (currentMethodInvocations == null) {
				currentMethodInvocations = new Vector<MethodInvocation>();
				methodToInvocationMap.put(methodInvocation.getIdentifier(), currentMethodInvocations);
			}
			currentMethodInvocations.add(methodInvocation);
		}
		return methodToInvocationMap;
	}	
	private static class MethodInvocationContext {
		private Map<Integer, Value> parameters;
		private String exceptionClass;
		public MethodInvocationContext(Map<Integer, Value> parameters, String exceptionClass) {
			super();
			this.parameters = parameters;
			this.exceptionClass = exceptionClass;
		}
		
		public boolean equals(Object o) {
			if (o != null && o instanceof MethodInvocationContext) {
				MethodInvocationContext peer = (MethodInvocationContext) o;
				if (peer.exceptionClass != null) {
					if (peer.exceptionClass.equals(this.exceptionClass)) {
						return peer.parameters.equals(this.parameters);
					} else {
						return false;
					}
				} else if (this.exceptionClass == null) {
					return peer.parameters.equals(this.parameters);
				} else {
					return false;
				}
			} else {
				return false;
			}
		}
		
		public int hashCode() {
			int result = parameters.hashCode();
			if (exceptionClass != null) {
				result += exceptionClass.hashCode();
			} else {
				result += 1;
			}
			return result;
		}
	}	
}
