package de.unisaarland.cs.st.tautoko.abstraction;

import java.io.IOException;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Vector;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.commons.EmptyVisitor;

import de.unisaarland.cs.st.adabu.util.asm.ASMUtilities;
import de.unisaarland.cs.st.adabu.util.asm.FieldIdentifier;
import de.unisaarland.cs.st.adabu.util.source.ClassFile;
import de.unisaarland.cs.st.adabu.util.source.ClassFileSource;
import de.unisaarland.cs.st.adabu.util.source.ClassPathClassFileSource;

public class EnumFinder implements Opcodes {

	public static List<FieldIdentifier> getNonStaticEnumFields(String className, ClassFileSource source) throws NoSuchElementException, IOException {
		final List<FieldIdentifier> allFields = new Vector<FieldIdentifier>();
		while (className != null && !className.equals("java/lang/Object")) {
			ClassFile classFile = source.getClassForName(className);
			ClassReader reader = new ClassReader(classFile.getClassBytes());
			final String theClassName = className;
			reader.accept(new ClassAdapter(new EmptyVisitor()) {
				@Override
				public FieldVisitor visitField(int access, String name, String desc, String signature, Object value) {
					FieldIdentifier identifier = new FieldIdentifier(theClassName, name, desc, ((access & ACC_STATIC) != 0));
					identifier.setAccess(access);
					allFields.add(identifier);
					return super.visitField(access, name, desc, signature, value);
				}
			}, ClassReader.SKIP_CODE + ClassReader.SKIP_DEBUG + ClassReader.SKIP_FRAMES );
			className = ASMUtilities.getSuperClassName(classFile.getClassBytes());
		}
		List<FieldIdentifier> enumFields = new Vector<FieldIdentifier>();
		for (FieldIdentifier fieldIdentifier : allFields) {
			if (!fieldIdentifier.isStatic() && fieldIdentifier.getMyType().isComplexType()) {
				String type = getClassNameFromType(fieldIdentifier.getType());
				if (isEnumType(type, source)) {
					enumFields.add(fieldIdentifier);
				}
			}
		}
		return enumFields;
	}

	public static String getClassNameFromType(String type) {
		return type.substring(1, type.length() - 1);
	}

	public static boolean isEnumType(String className, ClassFileSource source) throws IOException {
		try {
			String superClassName;
			do {
				ClassFile classFile = source.getClassForName(className);
				superClassName = ASMUtilities.getSuperClassName(classFile.getClassBytes());
				if (superClassName != null) {
					if (superClassName.equals("java/lang/Enum")) {
						return true;
					}
				} else {
					return false;
				}
				className = superClassName;
			} while (!superClassName.equals("java/lang/Object"));
			return false;
		} catch (NoSuchElementException exception) {
			return false;
		}
	}


	public static void main(String[] args) throws IOException {
		ClassPathClassFileSource source = new ClassPathClassFileSource(args[0]);
		List<FieldIdentifier> identifiers = getNonStaticEnumFields("de/unisaarland/cs/st/tautoko/abstraction/Test", source);
		for (FieldIdentifier fieldIdentifier : identifiers) {
			System.out.println(fieldIdentifier);
		}
	}

}
