package de.unisaarland.cs.st.tautoko.abstraction;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.Vector;

import javax.management.InstanceNotFoundException;

import org.apache.commons.io.FileUtils;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.util.TraceClassVisitor;

import de.unisaarland.cs.st.adabu.util.asm.ASMTreeUtilities;
import de.unisaarland.cs.st.adabu.util.asm.ASMUtilities;
import de.unisaarland.cs.st.adabu.util.asm.FieldIdentifier;
import de.unisaarland.cs.st.adabu.util.source.ClassFile;
import de.unisaarland.cs.st.adabu.util.source.ClassFileSource;

public class LimitedStateFieldFinder implements Opcodes {

	public static List<FieldIdentifier> getLimitedFields(String className, ClassFileSource source) throws NoSuchElementException, IOException {
		Map<String, FieldIdentifier> classFields = new HashMap<String, FieldIdentifier>();
		Set<FieldIdentifier> integerFields = new HashSet<FieldIdentifier>();
		Set<FieldIdentifier> usedIntegerFields = new HashSet<FieldIdentifier>();
		Set<FieldIdentifier> nonConstantFields = new HashSet<FieldIdentifier>();
		String currentClassName = className;
		while (currentClassName != null && !currentClassName.equals("java/lang/Object")) {
			ClassFile classFile = source.getClassForName(currentClassName);
			byte[] classBytes = classFile.getClassBytes();
			ClassNode classNode = ASMTreeUtilities.getClassNode(classBytes);
			for (Object object : classNode.fields) {
				FieldNode fieldNode = (FieldNode) object;
				FieldIdentifier fieldIdentifier = new FieldIdentifier(classNode.name, fieldNode.name, fieldNode.desc, (fieldNode.access & ACC_STATIC) != 0);
				fieldIdentifier.setAccess(fieldNode.access);
				classFields.put(fieldNode.name, fieldIdentifier);
				if (isNonPublicIntegerField(fieldNode)) {
					integerFields.add(fieldIdentifier);
				}
			}
			currentClassName = ASMUtilities.getSuperClassName(classBytes);
		}
		currentClassName = className;
		while (currentClassName != null && !currentClassName.equals("java/lang/Object")) {
			ClassFile classFile = source.getClassForName(currentClassName);
			byte[] classBytes = classFile.getClassBytes();
			ClassNode classNode = ASMTreeUtilities.getClassNode(classBytes);
			for (Object object : classNode.methods) {
				MethodNode methodNode = (MethodNode) object;
				InsnList instructions = methodNode.instructions;
				for (int i = 0; i < instructions.size(); i++) {
					AbstractInsnNode abstractNode = instructions.get(i);
					if (abstractNode instanceof FieldInsnNode) {
						FieldInsnNode fieldInstruction = (FieldInsnNode) abstractNode;
						if (fieldInstruction.getOpcode() == PUTFIELD) {
							FieldIdentifier identifier = new FieldIdentifier(fieldInstruction.owner, fieldInstruction.name, 
							                                                 fieldInstruction.desc, false);
							if (integerFields.contains(identifier)) {
								usedIntegerFields.add(identifier);
								if (!usesConstantValueForPut(instructions, fieldInstruction, classFields, classNode.name)) {
									nonConstantFields.add(identifier);
								}
							}
						}
					}
				}
				currentClassName = ASMUtilities.getSuperClassName(classBytes);
			}
		}
		usedIntegerFields.removeAll(nonConstantFields);
		return new Vector<FieldIdentifier>(usedIntegerFields);
	}

	private static boolean usesConstantValueForPut(InsnList instructions, FieldInsnNode fieldInstruction, Map<String, FieldIdentifier> allFields, String className) {
		AbstractInsnNode previousInstruction = fieldInstruction.getPrevious();
		if (previousInstruction instanceof LdcInsnNode) {
			return true;
		} else if (previousInstruction instanceof FieldInsnNode) {
			FieldInsnNode loadInstruction = (FieldInsnNode) previousInstruction;
			if (loadInstruction.owner.equals(className)) {
				if (allFields.containsKey(loadInstruction.name)) {
					FieldIdentifier identifier = allFields.get(loadInstruction.name);
					if ((identifier.getAccess() & ACC_FINAL) != 0) {
						return true;
					} else {
						return false;
					}
				} else {
					return false;
				}
			} else {
				return false;
			}
		} else if (previousInstruction instanceof InsnNode) {
			int opcode = previousInstruction.getOpcode();
			if (opcode == ICONST_0 ||
					opcode == ICONST_1 ||
					opcode == ICONST_2 ||
					opcode == ICONST_3 ||
					opcode == ICONST_4 ||
					opcode == ICONST_5) {
				return true;
			} else {
				return false;
			}
		} else {
			return false;
		}
	}

	private static void displayClass(byte[] classBytes) {
		ClassReader classReader;
		TraceClassVisitor traceVisitor;
		classReader = new ClassReader(classBytes);
		traceVisitor = new TraceClassVisitor(new PrintWriter(System.out));
		classReader.accept(traceVisitor, ClassReader.EXPAND_FRAMES);
	}   


	private static boolean isNonPublicIntegerField(FieldNode fieldNode) {
		return fieldNode.desc.equals("I") && ((fieldNode.access & ACC_PUBLIC) == 0) && ((fieldNode.access & ACC_FINAL) == 0) ;
	}

	public static void main(String[] args) throws IOException {
		byte[] bytes = FileUtils.readFileToByteArray(new File(args[0]));
		displayClass(bytes);
		/*List<FieldIdentifier> identifiers = getLimitedFields(bytes);
		for (FieldIdentifier fieldIdentifier : identifiers) {
			System.out.println(fieldIdentifier);
		}*/
	}
}
