package de.unisaarland.cs.st.tautoko.abstraction;

import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;

import de.unisaarland.cs.st.adabu.core.heappath.HeapPath;
import de.unisaarland.cs.st.adabu.core.state.Value;
import de.unisaarland.cs.st.adabu.trans.abstraction.StateAbstracter;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectModel;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectState;
import de.unisaarland.cs.st.adabu.util.abstraction.SimpleStateAbstracter;
import de.unisaarland.cs.st.adabu.util.asm.ASMUtilities;
import de.unisaarland.cs.st.adabu.util.asm.FieldIdentifier;
import de.unisaarland.cs.st.adabu.util.asm.MyType;
import de.unisaarland.cs.st.adabu.util.source.ClassFile;
import de.unisaarland.cs.st.adabu.util.source.ClassFileSource;

public class TypeStateStateAbstracter extends StateAbstracter {

	private ClassFileSource source;

	private List<FieldIdentifier> allConstantIntegerFields;

	private List<FieldIdentifier> allEnumFields;

	private Logger logger = Logger.getLogger(TypeStateStateAbstracter.class);

	private boolean useClassNameForComplexFields;

	public TypeStateStateAbstracter(TransitiveObjectModel model, ClassFileSource source) {
		this(model, source, false);
	}
	
	public TypeStateStateAbstracter(TransitiveObjectModel model, ClassFileSource source, boolean useClassNameForComplexFields) {
		super(model);
		this.source = source;
		this.allConstantIntegerFields = new Vector<FieldIdentifier>();
		this.allEnumFields = new Vector<FieldIdentifier>();
		this.useClassNameForComplexFields = useClassNameForComplexFields;
	}

	@Override
	protected TransitiveObjectState getAbstractState(TransitiveObjectState oldState) {
		return abstractState(oldState);
	}

	public TransitiveObjectState abstractState(TransitiveObjectState oldState) {
		TransitiveObjectState result = new TransitiveObjectState(oldState.getObjectId(), oldState.getTemporalOrderNumber());
		if (allConstantIntegerFields.size() > 0 || allEnumFields.size() > 0) {
			for (HeapPath path : oldState.getFieldValues().keySet()) {
				FieldIdentifier accessedField = path.getAccessedField();
				if (allEnumFields.contains(accessedField) || allConstantIntegerFields.contains(accessedField)) {
					Value value = oldState.getFieldValues().get(path);
					result.setFieldValue(path, value);
				}
			}
			return result;
		}
		for (HeapPath path : oldState.getFieldValues().keySet()) {
			FieldIdentifier accessedField = path.getAccessedField();
			Value value = oldState.getFieldValues().get(path);
			MyType accessedType = path.getAccessedType();
			if (accessedType.isComplexType()) {
				if (allEnumFields.contains(accessedField)) {
					result.setFieldValue(path, value);
				} else {
					if (!useClassNameForComplexFields || value.isNullReference()) {
						result.setFieldValue(path, SimpleStateAbstracter.getAbstractValue(value));
					} else {
						String typeString = value.getType().getTypeString();
						if (typeString.startsWith("L")) {
							result.setFieldValue(path, new Value(typeString.substring(1, typeString.length() - 1 ), value.getType()));
						} else {
							result.setFieldValue(path, SimpleStateAbstracter.getAbstractValue(value));
						}
					}
				}
			} else if (accessedType.isArrayType()) {
				result.setFieldValue(path, SimpleStateAbstracter.getAbstractValue(value));
			} else {
				if (accessedType.isIntegerType()) {
					if (allConstantIntegerFields.contains(accessedField)) {
						result.setFieldValue(path, value);
					} else {
						result.setFieldValue(path, SimpleStateAbstracter.getAbstractValue(value));
					}
				} else {
					result.setFieldValue(path, SimpleStateAbstracter.getAbstractValue(value));
				}
			}
		}
		return result;

	}

	@Override
	protected void prepare() {
		String className = model.getClassName();
		try {
			do {
				ClassFile classFile = source.getClassForName(className);
				List<FieldIdentifier> constantIntegerFields = LimitedStateFieldFinder.getLimitedFields(className, source);
				allConstantIntegerFields.addAll(constantIntegerFields);
				List<FieldIdentifier> enumFields = EnumFinder.getNonStaticEnumFields(className, source);
				allEnumFields.addAll(enumFields);
				String superClassName = ASMUtilities.getSuperClassName(classFile.getClassBytes());
				className = superClassName;
			} while (!className.equals("java/lang/Object"));
		} catch (Exception exception) {
			logger.warn("Exception determining constant or enum fields for class " + className, exception);
		}
	}

}
