package de.unisaarland.cs.st.tautoko.completion;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Vector;

import org.apache.commons.exec.CommandLine;
import org.apache.commons.exec.DefaultExecutor;
import org.apache.commons.exec.ExecuteException;
import org.apache.commons.exec.ExecuteWatchdog;
import org.apache.commons.exec.PumpStreamHandler;
import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;

public class FixRunner {

	protected static final Logger logger = Logger.getLogger(FixRunner.class);
	
	/*private GroovyShell shell;
	
	private HashMap<String, Script> scriptCache;*/

	protected List<JavaClass> fixes;

	protected File workingDirectory;

	protected String compileClassPath;

	protected File tmpDirectory;

	protected String testCommand;	
	
	public FixRunner(String testCommand, File workingDirectory, File tmpDirectory, List<JavaClass> fixes, String compileClassPath) {
		this.testCommand = testCommand;
		this.workingDirectory = workingDirectory;
		this.tmpDirectory = tmpDirectory;
		this.fixes = fixes;
		this.compileClassPath = compileClassPath;
		/*this.shell = new GroovyShell();
		this.scriptCache = new HashMap<String, Script>();*/
	}
	
	public List<JavaClass> processFixes() throws IOException {
		List<JavaClass> successfullFixes = new Vector<JavaClass>();
		int fixCounter = 1;
		for (JavaClass javaClass : fixes) {
			logger.debug("Checking fix: " + javaClass.getContent());
			compile(javaClass);
			boolean hasPassed = runTest(javaClass, fixCounter);
			if (hasPassed) {
				logger.info("Found a fix for the test: ");
				logger.info(javaClass.getContent());
				successfullFixes.add(javaClass);
			}
			fixCounter++;
		}
		return successfullFixes;
	}
	
	protected boolean runTest(JavaClass javaClass, int fixNumber) throws IOException {
		String modifierClass = javaClass.getPackageName() + "." + javaClass.getClassName();
		CommandLine commandLine = CommandLine.parse(testCommand);
		commandLine.addArgument("-Xbootclasspath/a:" + tmpDirectory.getAbsolutePath() + ":" + compileClassPath);
		commandLine.addArgument("-javaagent:/scratch2/dallmeier/bin/projects/adabu2/adabu2-pachika/target/adabu2-pachika-0.4-SNAPSHOT.jar");
		commandLine.addArgument("-Dadabu2.modifier=" + modifierClass);
		//commandLine.addArgument("-Dadabu2.display=true");
		return runProcess(commandLine, 0, workingDirectory);
	}

	protected void compile(JavaClass javaClass) throws IOException {
		File packageDirectory = new File(tmpDirectory.getAbsolutePath() + File.separator + javaClass.getPackageName().replace('.', '/'));
		packageDirectory.mkdirs();
		File classFile = new File(packageDirectory.getAbsolutePath() + File.separator + javaClass.getClassName() + ".java");
		FileUtils.writeStringToFile(classFile, javaClass.getContent(), null);
		CommandLine commandLine = null;
		if (System.getProperty("os.name").equalsIgnoreCase("linux")) {
			commandLine = CommandLine.parse("/usr/lib/jvm/jdk1.6.0/bin/javac");
		} else {
			commandLine = CommandLine.parse("/scratch2/dallmeier/soylatte/bin/javac");	
		}
		commandLine.addArguments("-cp");
		commandLine.addArgument(compileClassPath + ":.");
		commandLine.addArgument(javaClass.getPackageName().replace('.', '/') + "/" + javaClass.getClassName() + ".java");
		runProcess(commandLine, 0, tmpDirectory);
	}
	
	protected boolean runProcess(CommandLine commandLine, int expectedExitValue, File workingDirectory) throws IOException {
		logger.debug("Executing command line " + commandLine.toString());
		ExecuteWatchdog watchdog = new ExecuteWatchdog(60000);
		DefaultExecutor executor = new DefaultExecutor();
		logger.debug("Setting working directory " + workingDirectory + ".");
		executor.setWorkingDirectory(workingDirectory);
		executor.setWatchdog(watchdog);
		ByteArrayOutputStream out = new ByteArrayOutputStream(); 
		executor.setExitValue(expectedExitValue);
		executor.setStreamHandler(new PumpStreamHandler(out));
		
		try {
			int exitValue = executor.execute(commandLine);		
			String output = out.toString();
			try {
			  Thread.sleep(500);
			} catch (InterruptedException exception) {
				//ignore exception
			}
			if (output != null) {
				logger.debug("Process exit value " + exitValue);
				logger.trace("Output " + output);
			}
			return true;
		} catch (ExecuteException exception) {
			int exitValue = exception.getExitValue();
			String output = out.toString();
			logger.error("Process threw an exception and ended with " + exitValue, exception);
			logger.error("Output " + output);
			return false;
		}
	}
	
	/*private String runGroovyScript(String script, Binding binding) {
		Script groovyScript = scriptCache.get(script);
		if (groovyScript == null) {
			groovyScript = shell.parse(script);
			scriptCache.put(script, groovyScript);
		}
		groovyScript.setBinding(binding);
		try {
		  Object value = groovyScript.run();
		  return binding.getVariable("result").toString();
		} catch (Throwable executionException) {
			logger.error("An exception occured executing the script", executionException);
			return null;
		}
	}*/
}
