package de.unisaarland.cs.st.tautoko.instrument;

import java.io.PrintWriter;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.util.TraceClassVisitor;

public class Instrumenter implements ClassFileTransformer {

	public static final String PROP_CLASSNAME_MODIFIER = "adabu.modifier";

	public static final String PROP_DISPLAY = "adabu.display";
	
	public static final String PROP_DEBUG = "adabu.debug";

	private static Instrumenter singletonInstance;

	private static boolean debug = false;
	
	private IModifier modifier;

	private Instrumenter(IModifier modifier) {
		this.modifier = modifier;
	}

	public static void premain(String agentArgs, Instrumentation inst) {
		// check the JRE version we run on
		final String javaVersion = System.getProperty("java.version");
		final int secondPointPos = javaVersion.indexOf('.', javaVersion.indexOf('.')+1);
		try {
			if (secondPointPos != -1) {
				final double javaVersionDouble = Double.valueOf(javaVersion.substring(0, secondPointPos));
				if (javaVersionDouble < 1.59) {
					System.out.println("This tracer requires JRE >= 1.6, you are running " + javaVersion + ".");
					System.exit(-1);
				}
			}
		} catch (final NumberFormatException e) {
			// ignore (no check...)
		}		
		if (!System.getProperties().containsKey(PROP_CLASSNAME_MODIFIER)) {
			System.out.println("Property " + PROP_CLASSNAME_MODIFIER + " not set. Please specifiy the fully qualified classname of a modifier.");
			System.exit(-1);
		} else {
			try {
				if (System.getProperty(PROP_DEBUG) != null && System.getProperty(PROP_DEBUG).equals("true")) {
					debug = true;
				}
				Class<?> modifierClass = Class.forName(System.getProperty(PROP_CLASSNAME_MODIFIER));
				IModifier modifier = (IModifier) modifierClass.newInstance();
				modifier.init();
				singletonInstance = new Instrumenter(modifier);
				inst.addTransformer(singletonInstance);
			} catch (Exception loadException) {
				System.out.println("An exception occured loading the modifier class:");
				System.out.println(loadException);
				loadException.printStackTrace();
				System.exit(-1);
			}
		}
	}

	private byte[] changeClassVisibilityToPublic(String className, byte[] classBytes) {
		ClassReader classReader = new ClassReader(classBytes);
		ClassWriter classWriter = new ClassWriter(ClassWriter.COMPUTE_MAXS);
		ClassVisibilityClassAdapter adapter = new ClassVisibilityClassAdapter(classWriter);
		classReader.accept(adapter, 0);
		return classWriter.toByteArray();
	}

	public byte[] transform(ClassLoader loader, String className, Class<?> classBeingRedefined,
	                        ProtectionDomain protectionDomain, byte[] classBytes) throws IllegalClassFormatException {
		try {
			byte[] modifiedBytes = modifier.transform(className, classBytes);
			if (modifiedBytes != null) {
				if (System.getProperty(PROP_DISPLAY) != null) {
					displayClass(modifiedBytes);
				}
			} else {
				modifiedBytes = changeClassVisibilityToPublic(className, classBytes);
			}
			return modifiedBytes;
		} catch (Throwable instrumentationException) {
			System.out.println("An exception occured instrumenting: " + instrumentationException);
			instrumentationException.printStackTrace();
			try {
				Thread.sleep(500);
			} catch (InterruptedException exc) {}
			System.exit(-1);
			return null;
		}
	}

	private static void displayClass(byte[] classBytes) {
		ClassReader classReader;
		TraceClassVisitor traceVisitor;

		classReader = new ClassReader(classBytes);
		traceVisitor = new TraceClassVisitor(new PrintWriter(System.out));
		classReader.accept(traceVisitor, 0);
	}   

}
