package de.unisaarland.cs.st.tautoko.miner;

import java.io.File;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import de.unisaarland.cs.st.adabu.core.AbstractTraceFileReader;
import de.unisaarland.cs.st.adabu.core.CallStack;
import de.unisaarland.cs.st.adabu.core.InvokedMethod;
import de.unisaarland.cs.st.adabu.core.StateKeepingTraceFileReader;
import de.unisaarland.cs.st.adabu.core.heappath.FieldAccess;
import de.unisaarland.cs.st.adabu.core.heappath.HeapPath;
import de.unisaarland.cs.st.adabu.core.state.ObjectState;
import de.unisaarland.cs.st.adabu.core.state.Value;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocation;
import de.unisaarland.cs.st.adabu.trans.model.ObjectField;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectModel;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectState;
import de.unisaarland.cs.st.adabu.util.ObjectPool;
import de.unisaarland.cs.st.adabu.util.asm.FieldIdentifier;
import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.adabu.util.asm.MyType;
import de.unisaarland.cs.st.adabu.util.collections.IntHashSet;
import de.unisaarland.cs.st.adabu.util.collections.IntIterator;
import de.unisaarland.cs.st.adabu.util.collections.IntHashSet.MyIntIterator;

public abstract class TypeStateMiner extends StateKeepingTraceFileReader {

	private Logger logger = Logger.getLogger(TypeStateMiner.class);

	private int maxHeapPathLength;

	private HashMap<Integer, Stack<HashMap<ObjectField, Value>>> threadIdToFieldsReadMap;

	private HashMap<Integer, Stack<HashMap<ObjectField, Value>>> threadIdToFieldsWrittenMap;

	private HashMap<Integer, Stack<HashMap<ObjectField, Value>>> threadIdToOldValuesMap;

	private HashMap<Integer, Stack<MethodIdentifier>> threadIdToFinalizerMethodStackMap;

	private ObjectPool<HashMap<ObjectField, Value>> fieldMapPool = new ObjectPool<HashMap<ObjectField, Value>>();

	private ObjectPool<IntHashSet> intHashSetPool = new ObjectPool<IntHashSet>();

	private ObjectPool<MyIntIterator> iteratorPool = new ObjectPool<MyIntIterator>();

	private HashMap<Integer, Stack<IntHashSet>> threadIdToCreatedObjectStackMap;

	private HashMap<Integer, TransitiveObjectModel> objectIdToModelMap;

	private HashMap<Integer, TransitiveObjectState> objectIdToTransitiveStateMap;

	private Pattern finalizerMethodPattern;


	public TypeStateMiner(File traceFile, int maxHeapPathLength) {
		this(traceFile, maxHeapPathLength, null);
	}

	public TypeStateMiner(File traceFile, int maxHeapPathLength, Pattern finalizerMethodPattern) {
		super(traceFile);
		this.setFinalizerMethodPattern(finalizerMethodPattern);
		this.maxHeapPathLength = maxHeapPathLength;
		this.threadIdToFieldsReadMap = new HashMap<Integer, Stack<HashMap<ObjectField,Value>>>();
		this.threadIdToFieldsWrittenMap = new HashMap<Integer, Stack<HashMap<ObjectField,Value>>>();
		this.threadIdToOldValuesMap = new HashMap<Integer, Stack<HashMap<ObjectField,Value>>>();
		this.threadIdToFinalizerMethodStackMap = new HashMap<Integer, Stack<MethodIdentifier>>();
		this.threadIdToCreatedObjectStackMap = new HashMap<Integer, Stack<IntHashSet>>();
		this.objectIdToModelMap = new HashMap<Integer, TransitiveObjectModel>();
		this.objectIdToTransitiveStateMap = new HashMap<Integer, TransitiveObjectState>();
	}


	@Override
	protected void processArrayRead(Integer threadId, int index, Integer arrayId,
	                                FieldIdentifier identifier, Value value) {
		super.processArrayRead(threadId, index, arrayId, identifier, value);
		HashMap<ObjectField, Value> fieldsRead = getOrCreateStack(threadId, threadIdToFieldsReadMap).peek();
		addToMap(fieldsRead, arrayId, identifier, value, false);
	}


	@Override
	protected void processArrayWrite(Integer threadId, int index,
	                                 Integer arrayId,
	                                 FieldIdentifier identifier, Value value) {
		rescueOldValue(threadId, identifier, arrayId);
		super.processArrayWrite(threadId, index, arrayId, identifier, value);
		HashMap<ObjectField, Value> fieldsWritten = getOrCreateStack(threadId, threadIdToFieldsWrittenMap).peek();
		addToMap(fieldsWritten , arrayId, identifier, value, true);
	}


	@Override
	protected void processFieldRead(Integer threadId, FieldIdentifier identifier,
	                                Integer objectId, Value value) {
		super.processFieldRead(threadId, identifier, objectId, value);
		HashMap<ObjectField, Value> fieldsRead = getOrCreateStack(threadId, threadIdToFieldsReadMap).peek();
		addToMap(fieldsRead, objectId, identifier, value, false);
	}


	@Override
	protected void handleFieldWriteObjectEvent(Integer threadId, Integer targetObjectId,
	                                           FieldIdentifier identifier, int value, 
	                                           String className) {
		rescueOldValue(threadId, identifier, targetObjectId);
		super.handleFieldWriteObjectEvent(threadId, targetObjectId, identifier, value, className);
	}

	@Override
	protected void processFieldWrite(Integer threadId,
	                                 FieldIdentifier identifier,
	                                 Integer objectId, Value value) {
		if (!identifier.isComplexType()) {
			rescueOldValue(threadId, identifier, objectId);
		}
		super.processFieldWrite(threadId, identifier, objectId, value);
		HashMap<ObjectField, Value> fieldsWritten = getOrCreateStack(threadId, threadIdToFieldsWrittenMap).peek();
		addToMap(fieldsWritten , objectId, identifier, value, true);
	}

	private void rescueOldValue(Integer threadId, FieldIdentifier identifier, Integer objectId) {
		ObjectState state = getObjectState(objectId);
		if (state != null) {
			Value oldValue = state.getValue(identifier);
			ObjectField field = new ObjectField(objectId, identifier);
			HashMap<ObjectField, Value> oldValueMap = threadIdToOldValuesMap.get(threadId).peek();
			if (!oldValueMap.containsKey(field)) {
				oldValueMap.put(field, oldValue);
			}
		}
	}

	private void addToMap(HashMap<ObjectField, Value> map, Integer objectId, FieldIdentifier identifier, Value value, boolean overwrite) {
		ObjectField objectField = new ObjectField(objectId, identifier);
		if (overwrite || !map.containsKey(objectField)) {
			map.put(objectField, value);
		}
	}

	protected abstract boolean isInterestingObject(Integer objectId);

	protected String getClassNameFromType(MyType type) {
		assert (type.isComplexType());
		return type.getTypeString().substring(1, type.getTypeString().length() - 1);
	}

	@SuppressWarnings("unchecked")
	private MethodInvocation getInvocation(MethodIdentifier identifier, CallStack callStack, HashSet<HeapPath> preconditions, 
	                                       HashMap<Integer, Value> parameters, int callSiteId, String exceptionClass) {
		MethodInvocation result = new MethodInvocation(identifier, preconditions, new HashSet<HeapPath>(), callStack);
		result.setExceptionClass(exceptionClass);
		result.setParameters((Map<Integer, Value>) parameters.clone());
		result.setCallSiteId(callSiteId);
		return result;
	}

	protected void updateModelsForObjects(InvokedMethod activeMethod, IntHashSet changedObjects, 
	                                      MethodIdentifier identifier, CallStack callStack, 
	                                      HashMap<ObjectField, Value> fieldsRead, HashMap<Integer, Value> parameters, String exceptionClass) {
		MyIntIterator iterator = getCachedOrNewIterator();
		changedObjects.iterator(iterator);
		HashSet<HeapPath> preconditions = null;
		while (iterator.hasNext()) {
			Integer objectId = iterator.nextInt();
			if (isInterestingObject(objectId)) {
				ObjectState state = getObjectState(objectId);
				if (state != null) {
					if (state.isArray()) {
						logger.debug("Ignoring changes for object " + objectId + " because it's an array.");
						continue;
					}
					String className = getClassNameFromType(state.getType());
					TransitiveObjectModel model = objectIdToModelMap.get(objectId);
					//we ignore method calls that happen before the first constructor call has finished
					if (model == null && !identifier.isConstructor()) {
						continue;
					}
					if (model == null) {
						model = new TransitiveObjectModel(className, objectId);
						objectIdToModelMap.put(objectId, model);
					}
					TransitiveObjectState newState = getDeepStateForObject(objectId);
					TransitiveObjectState oldState = objectIdToTransitiveStateMap.get(objectId);
					if (oldState == null) {
						oldState = new TransitiveObjectState(objectId);
						model.setStartingState(oldState);
						model.setCurrentState(oldState);
					}
					CallStack clonedStack = callStack.clone();
					clonedStack.push(activeMethod);
					if (preconditions == null) {
						preconditions = getHeapPaths(activeMethod.getThisObjectId(), fieldsRead);
					}
					MethodInvocation invocation = getInvocation(identifier, clonedStack, preconditions, parameters, activeMethod.getCallSiteId(), exceptionClass);
					if (!objectId.equals(activeMethod.getThisObjectId())) {
						HeapPath path = getHeapPathForChangedObject(changedObjects, objectId, activeMethod.getThisObjectId());
						if (path != null) {
							invocation.setAccessedObjectPath(path);
						}
					}
					model.addTransition(oldState, newState, invocation);
					//System.out.println("New model for object " + objectId + " after call to " + identifier + ": "  + model);
					objectIdToTransitiveStateMap.put(objectId, newState);
				}
			}
		}
		returnIteratorToPool(iterator);
		int thisObjectId = activeMethod.getThisObjectId();
		if (isInterestingObject(thisObjectId) && !changedObjects.contains(thisObjectId)) {
			ObjectState objectState = getObjectState(thisObjectId);
			if (objectState != null) {
				CallStack clonedStack = callStack.clone();
				clonedStack.push(activeMethod);
				if (preconditions == null) {
					preconditions = getHeapPaths(activeMethod.getThisObjectId(), fieldsRead);
				}
				MethodInvocation invocation = getInvocation(identifier, clonedStack, preconditions, parameters, activeMethod.getCallSiteId(), exceptionClass);
				TransitiveObjectModel model = objectIdToModelMap.get(thisObjectId);
				if (model == null && activeMethod.getIdentifier().isConstructor()) {
					String className = getClassNameFromType(objectState.getType());
					model = new TransitiveObjectModel(className, thisObjectId);
					TransitiveObjectState startingState = new TransitiveObjectState(thisObjectId);
					model.setStartingState(startingState);
					model.setCurrentState(startingState);
					objectIdToModelMap.put(thisObjectId, model);
				} 
				//it sometimes happens that models are still null
				//I think this has to do with reflection
				if (model != null) {
					model.addTransition(model.getCurrentState(), model.getCurrentState(), invocation);
				}
			}   
		}
	}


	private HashSet<HeapPath> getHeapPaths(int thisObjectId, HashMap<ObjectField, Value> fieldsRead) {
		HashMap<Integer, ObjectField> idToFieldMap = new HashMap<Integer, ObjectField>();
		for (ObjectField field : fieldsRead.keySet()) {
			idToFieldMap.put(field.getObjectId(), field);
		}
		HashSet<HeapPath> allPaths = new HashSet<HeapPath>(fieldsRead.size());
		ObjectState startObjectState = getObjectState(thisObjectId);
		Set<FieldIdentifier> complexFields = startObjectState.getComplexFields();
		for (FieldIdentifier fieldIdentifier : complexFields) {
			Value value = startObjectState.getValue(fieldIdentifier);
			if (!value.isNullReference())   {
				Integer currentObjectId = value.getObjectId();
				FieldAccess fieldAccess = new FieldAccess(null, startObjectState.getObjectId(), fieldIdentifier, false);
				HeapPath path = new HeapPath(fieldAccess);
				if (idToFieldMap.containsKey(currentObjectId)) {
					HeapPath completePath = (HeapPath) path.clone();
					FieldAccess finalFieldAccess = new FieldAccess(null, currentObjectId, idToFieldMap.get(currentObjectId).getIdentifier(), false);
					completePath.appendElement(finalFieldAccess);
					allPaths.add(completePath);
					idToFieldMap.remove(currentObjectId);
				}
				if (maxHeapPathLength > 0) {
					addPathsRecursively(value.getObjectId(), path, allPaths, idToFieldMap, maxHeapPathLength - 1);
				} 
			}
		}
		return allPaths;
	}

	protected void addPathsRecursively(int objectId, HeapPath pathStart, Set<HeapPath> allPaths, HashMap<Integer, ObjectField> idToFieldMap, int maxHeapPathLength) {
		ObjectState startObjectState = getObjectState(objectId);
		if (startObjectState != null) {
			Set<FieldIdentifier> complexFields = startObjectState.getComplexFields();
			for (FieldIdentifier fieldIdentifier : complexFields) {
				Value value = startObjectState.getValue(fieldIdentifier);
				if (!value.isNullReference())   {
					Integer currentObjectId = value.getObjectId();
					FieldAccess fieldAccess = new FieldAccess(null, startObjectState.getObjectId(), fieldIdentifier, false);
					HeapPath path = (HeapPath) pathStart.clone();
					path.appendElement(fieldAccess);
					if (idToFieldMap.containsKey(currentObjectId)) {
						HeapPath completePath = (HeapPath) path.clone();
						FieldAccess finalFieldAccess = new FieldAccess(null, currentObjectId, idToFieldMap.get(currentObjectId).getIdentifier(), false);
						completePath.appendElement(finalFieldAccess);
						allPaths.add(completePath);
						idToFieldMap.remove(currentObjectId);
					}   
					if (maxHeapPathLength > 0) {
						addPathsRecursively(value.getObjectId(), path, allPaths, idToFieldMap, maxHeapPathLength - 1);
					} 
				}
			}
		} else {
			logger.trace("Unable to add paths because there is no state information for object " + objectId);
		}
	}

	protected void addPathsRecursively(Integer fromObjectId, HeapPath pathStart, IntHashSet changedObjects, 
	                                   Integer toObjectId, int maxHeapLength, List<HeapPath> paths) {
		ObjectState objectState = getObjectState(fromObjectId);
		if (objectState != null) {
			Set<FieldIdentifier> complexFields = objectState.getComplexFields();
			if (fromObjectId.equals(toObjectId)) {
				paths.add(pathStart);
			} else {
				if (maxHeapLength > 0) {
					for (FieldIdentifier fieldIdentifier : complexFields) {
						Value value = objectState.getValue(fieldIdentifier);
						if (!value.isNullReference())   {
							if (changedObjects.contains(value.getObjectId())) {
								FieldAccess fieldAccess = new FieldAccess(null, objectState.getObjectId(), fieldIdentifier, false);
								HeapPath path = (HeapPath) pathStart.clone();
								path.appendElement(fieldAccess);
								addPathsRecursively(value.getObjectId(), path, changedObjects, toObjectId, maxHeapLength - 1, paths);
							}
						}
					}
				}
			}
		}
	}

	protected HeapPath getHeapPathForChangedObject(IntHashSet changedObjects, Integer fromObjectId, Integer toObjectId) {
		List<HeapPath> paths = new Vector<HeapPath>();
		ObjectState startObjectState = getObjectState(fromObjectId);
		Set<FieldIdentifier> complexFields = startObjectState.getComplexFields();
		for (FieldIdentifier fieldIdentifier : complexFields) {
			Value value = startObjectState.getValue(fieldIdentifier);
			if (!value.isNullReference())   {
				if (changedObjects.contains(value.getObjectId())) {
					FieldAccess fieldAccess = new FieldAccess(null, startObjectState.getObjectId(), fieldIdentifier, false);
					HeapPath path = new HeapPath(fieldAccess);
					addPathsRecursively(value.getObjectId(), path, changedObjects, toObjectId, this.maxHeapPathLength - 1, paths);
				}
			}
		}
		Collections.sort(paths, new Comparator<HeapPath>() {
			public int compare(HeapPath o1, HeapPath o2) {
				return o1.size() - o2.size();
			}

		});
		if (!paths.isEmpty()) {
			return paths.get(0);
		} else {
			return null;
		}
	}

	protected TransitiveObjectState getDeepStateForObject(Integer objectId) {
		TransitiveObjectState newState = new TransitiveObjectState(objectId);
		ObjectState state = getObjectState(objectId);
		int heapDepth = maxHeapPathLength;
		if (state != null) {
			for (FieldIdentifier identifier : state.getFieldValues().keySet()) {
				FieldAccess fieldAccess = new FieldAccess(null, objectId, identifier, false);
				HeapPath path = new HeapPath(fieldAccess);
				Value value = state.getValue(identifier);
				newState.setFieldValue(path, value);
				MyType type = value.getType();
				if (heapDepth > 0) {
					if (type.isComplexType() || type.isArrayType()) {
						Integer recursiveObjectId = value.getObjectId();
						addStateRecursively(newState, recursiveObjectId, path, heapDepth - 1);
					}
				}
			}
		}
		return newState;
	}

	protected void addStateRecursively(TransitiveObjectState newState, Integer objectId, HeapPath pathStart, int heapDepth) {
		ObjectState state = getObjectState(objectId);
		if (state != null) {
			for (FieldIdentifier identifier : state.getFieldValues().keySet()) {
				FieldAccess fieldAccess = new FieldAccess(null, objectId, identifier, false);
				HeapPath path = (HeapPath) pathStart.clone();
				path.appendElement(fieldAccess);
				Value value = state.getValue(identifier);
				newState.setFieldValue(path, value);
				MyType type = value.getType();
				if (heapDepth > 0) {
					if (type.isComplexType() || type.isArrayType()) {
						Integer recursiveObjectId = value.getObjectId();
						addStateRecursively(newState, recursiveObjectId, path, heapDepth - 1);
					}
				}
			}
		}
	}

	@Override
	protected void handleMethodEndEvent(Integer threadId,
	                                    MethodIdentifier identifier, 
	                                    String exceptionClass) {
		InvokedMethod activeMethod = getActiveMethod(threadId);
		CallStack callStack = getOrCreateMethodStackForThread(threadId);
		HashMap<Integer, Value> parameters = threadIdToParameterSetStackMap.get(threadId).peek();
		Value returnedValue = threadIdToReturnedValueMap.remove(threadId);
		IntHashSet visibleObjects = getVisibleObjects(parameters, returnedValue);
		IntHashSet createdObjects = threadIdToCreatedObjectStackMap.get(threadId).pop();
		super.handleMethodEndEvent(threadId, identifier, exceptionClass);
		HashMap<ObjectField, Value> fieldsRead = threadIdToFieldsReadMap.get(threadId).pop();
		HashMap<ObjectField, Value> fieldsWritten = threadIdToFieldsWrittenMap.get(threadId).pop();
		HashMap<ObjectField, Value> oldValues = threadIdToOldValuesMap.get(threadId).pop();
		//for now only include direct changes
		//IntHashSet changedObjects = calculateChangedObjects(fieldsWritten, oldValues);
		IntHashSet changedObjects = new IntHashSet();
		if (!activeMethod.getIdentifier().isStatic()) {
			changedObjects.add(activeMethod.getThisObjectId());
		}
		if (!isFinalizerMethodActive(threadId) && !isInstanceMethodOnTop(threadId, activeMethod.getThisObjectId())) {
			updateModelsForObjects(activeMethod, changedObjects, identifier, callStack, fieldsRead, parameters, exceptionClass);
		} 
		propagateFieldsReadInformation(threadId, fieldsRead);
		//run garbage collection
		cleanUp(visibleObjects, createdObjects);
		//return objects into pools
		returnMapToPool(fieldsRead);
		returnMapToPool(fieldsWritten);
		returnMapToPool(oldValues);
		returnSetToPool(createdObjects);
		returnSetToPool(visibleObjects);
		returnSetToPool(changedObjects);
		if (isFinalizerMethod(identifier)) {
			getOrCreateFinalizerStackForThread(threadId).pop();
		}
	}

	protected boolean isInstanceMethodOnTop(Integer threadId, Integer objectId) {
		CallStack callStack = getOrCreateMethodStackForThread(threadId);
		if (callStack.size() > 0) {
			InvokedMethod activeMethod = callStack.getActiveMethod();
			return activeMethod.getThisObjectId() == objectId;
		} else {
			return false;
		}
	}

	protected void propagateFieldsReadInformation(Integer threadId, HashMap<ObjectField, Value> fieldsRead) {
		InvokedMethod activeMethod = getActiveMethod(threadId);
		if (activeMethod != null && !activeMethod.getIdentifier().isStatic()) {
			int thisObjectId = activeMethod.getThisObjectId();
			HashMap<ObjectField, Value> callingMethodFieldsRead = threadIdToFieldsReadMap.get(threadId).peek();
			IntHashSet createdObjects = threadIdToCreatedObjectStackMap.get(threadId).peek();
			IntHashSet referencedObjects = getReferencedObjects(thisObjectId, maxHeapPathLength);
			for (ObjectField field : fieldsRead.keySet()) {
				int accessedObjectId = field.getObjectId();
				if (!createdObjects.contains(accessedObjectId) && referencedObjects.contains(accessedObjectId) && !callingMethodFieldsRead.containsKey(field)) {
					callingMethodFieldsRead.put(field, fieldsRead.get(field));
				}
			}
		}
	}

	protected boolean isFinalizerMethodActive(Integer threadId) {
		return getOrCreateFinalizerStackForThread(threadId).size() > 0;
	}

	protected boolean isFinalizerMethod(MethodIdentifier identifier) {
		if (finalizerMethodPattern == null) {
			return false;
		} else {
			Matcher matcher = finalizerMethodPattern.matcher(identifier.toString());
			return matcher.matches();
		}
	}

	protected IntHashSet calculateChangedObjects(HashMap<ObjectField, Value> fieldsWritten, HashMap<ObjectField, Value> oldValues) {
		IntHashSet changedObjects = getCachedOrNewSet();
		//we don't need to take care of old values for complex fields,
		//since the referencing object must also have changed
		//a.b = c changed to
		//a.b = d and
		//d.e = f then we know that a also must have changed.
		for (ObjectField field : fieldsWritten.keySet()) {
			Value newValue = fieldsWritten.get(field);
			Value oldValue = oldValues.get(field);
			if (oldValue == null || !newValue.getData().equals(oldValue.getData())) {
				changedObjects.add(field.getObjectId());
			}
		}
		IntHashSet result = getCachedOrNewSet();
		addAll(changedObjects, result);
		MyIntIterator iterator = getCachedOrNewIterator();
		changedObjects.iterator(iterator);
		while (iterator.hasNext()) {
			int objectId = iterator.nextInt();

			IntHashSet referencingObjects = getReferencingObjects(objectId, maxHeapPathLength);
			addAll(referencingObjects, result);
			returnSetToPool(referencingObjects);
		}
		returnIteratorToPool(iterator);
		returnSetToPool(changedObjects);
		return result;
	}

	protected void addAll(IntHashSet from, IntHashSet to) {
		MyIntIterator iterator = getCachedOrNewIterator();
		from.iterator(iterator);
		while (iterator.hasNext()) {
			to.add(iterator.nextInt());
		}
		returnIteratorToPool(iterator);
	}

	protected IntHashSet getReferencedObjects(Integer objectId, int heapDepth) {
		IntHashSet allReferencedObjects = getCachedOrNewSet();
		IntHashSet referencedObjects = objectIdToPointsToMap.get(objectId);
		if (referencedObjects != null) {
			MyIntIterator iterator = getCachedOrNewIterator();
			referencedObjects.iterator(iterator);
			while (iterator.hasNext()) {
				int referencedObjectId = iterator.nextInt();
				allReferencedObjects.add(referencedObjectId);
				if (heapDepth > 1) {
					IntHashSet recursiveReferencedObjects = getReferencedObjects(referencedObjectId, heapDepth - 1);
					addAll(recursiveReferencedObjects, allReferencedObjects);
					returnSetToPool(recursiveReferencedObjects);
				}
			}
			returnIteratorToPool(iterator);
		}
		return allReferencedObjects;
	}

	protected IntHashSet getReferencingObjects(Integer objectId, int heapDepth) {
		IntHashSet changedObjects = getCachedOrNewSet();
		IntHashSet referencingObjects = objectIdToReferencedByMap.get(objectId);
		if (referencingObjects != null) {
			MyIntIterator iterator = getCachedOrNewIterator();
			referencingObjects.iterator(iterator);
			while (iterator.hasNext()) {
				int referencingObjectId = iterator.nextInt();
				changedObjects.add(referencingObjectId);
				if (heapDepth > 1) {
					IntHashSet recursiveReferencingObjects = getReferencingObjects(referencingObjectId, heapDepth - 1);
					addAll(recursiveReferencingObjects, changedObjects);
					returnSetToPool(recursiveReferencingObjects);
				}
			}
			returnIteratorToPool(iterator);
		}
		return changedObjects;
	}

	protected void cleanUp(IntHashSet visibleObjects, IntHashSet createdObjects) {
		IntIterator iterator = (IntIterator) createdObjects.iterator();
		while (iterator.hasNext()) {
			int createdObjectId = iterator.nextInt();
			if (!visibleObjects.contains(createdObjectId)) {
				removeObject(createdObjectId);
			}
		}
	}

	protected IntHashSet getVisibleObjects(HashMap<Integer, Value> parameters,
	                                       Value returnedValue) {
		IntHashSet result = getCachedOrNewSet();
		for (Integer index: parameters.keySet()) {
			Value parameterValue = parameters.get(index);
			if (parameterValue.getType().isArrayType() || parameterValue.getType().isComplexType()) {
				result.add(parameterValue.getObjectId().intValue());
				addReferencedObjectsToSet(parameterValue.getObjectId(), result);
			}
		}
		if (returnedValue != null) {
			MyType valueType = returnedValue.getType();
			if (valueType.isComplexType() || valueType.isArrayType()) {
				int returnedObjectId = returnedValue.getObjectId();
				if (!result.contains(returnedObjectId)) {
					result.add(returnedObjectId);
					addReferencedObjectsToSet(returnedObjectId, result);
				}
			}
		}
		result.add(AbstractTraceFileReader.ID_STATIC);
		addReferencedObjectsToSet(AbstractTraceFileReader.ID_STATIC, result);
		return result;
	}

	@Override
	protected void handleMethodStartEvent(Integer threadId,
	                                      MethodIdentifier identifier,
	                                      int thisObjectId) {
		super.handleMethodStartEvent(threadId, identifier, thisObjectId);
		if (isFinalizerMethod(identifier)) {
			getOrCreateFinalizerStackForThread(threadId).push(identifier);
		}
		getOrCreateStack(threadId, threadIdToFieldsReadMap).push(getCachedOrNewMap());
		getOrCreateStack(threadId, threadIdToFieldsWrittenMap).push(getCachedOrNewMap());
		getOrCreateStack(threadId, threadIdToOldValuesMap).push(getCachedOrNewMap());
		getOrCreateObjectStackMapForThread(threadId).push(getCachedOrNewSet());
	}


	private Stack<HashMap<ObjectField, Value>> getOrCreateStack(Integer threadId, 
	                                                            HashMap<Integer, Stack<HashMap<ObjectField, Value>>> map) {
		Stack<HashMap<ObjectField, Value>> stack = map.get(threadId);
		if (stack == null) {
			stack = new Stack<HashMap<ObjectField,Value>>();
			map.put(threadId, stack);
		}
		return stack;
	}

	@Override
	protected void handleArrayCreatedEvent(Integer threadId, int newArrayId,
	                                       String typeDesc) {
		super.handleArrayCreatedEvent(threadId, newArrayId, typeDesc);
		threadIdToCreatedObjectStackMap.get(threadId).peek().add(newArrayId);
	}

	@Override
	protected void handleObjectCreatedEvent(Integer threadId, Integer newObjectId,
	                                        String className, int newSiteId) {
		super.handleObjectCreatedEvent(threadId, newObjectId, className, newSiteId);
		threadIdToCreatedObjectStackMap.get(threadId).peek().add(newObjectId.intValue());
	}

	protected Stack<IntHashSet> getOrCreateObjectStackMapForThread(Integer threadId) {
		Stack<IntHashSet> result = threadIdToCreatedObjectStackMap.get(threadId);
		if (result == null) {
			result = new Stack<IntHashSet>();
			threadIdToCreatedObjectStackMap.put(threadId, result);
		}
		return result;
	}

	private void returnMapToPool(HashMap<ObjectField, Value> map) {
		map.clear();
		fieldMapPool.addObject(map);
	}

	private HashMap<ObjectField, Value> getCachedOrNewMap() {
		if (fieldMapPool.size() > 0) {
			return fieldMapPool.getObject();
		} else {
			return new HashMap<ObjectField, Value>(20);
		}
	}

	private void returnSetToPool(IntHashSet set) {
		set.clear();
		intHashSetPool.addObject(set);
	}

	private IntHashSet getCachedOrNewSet () {
		if (intHashSetPool.size() > 0) {
			return intHashSetPool.getObject();
		} else {
			return new IntHashSet(20);
		}
	}

	private void returnIteratorToPool(MyIntIterator iterator) {
		iteratorPool.addObject(iterator);
	}

	private MyIntIterator getCachedOrNewIterator() {
		if (iteratorPool.size() > 0) {
			return iteratorPool.getObject();
		} else {
			return (MyIntIterator) new IntHashSet().iterator();
		}
	}


	public List<TransitiveObjectModel> getModels() {
		return new Vector<TransitiveObjectModel>(objectIdToModelMap.values());
	}


	public void setFinalizerMethodPattern(Pattern finalizerMethodPattern) {
		this.finalizerMethodPattern = finalizerMethodPattern;
	}


	public Pattern getFinalizerMethodPattern() {
		return finalizerMethodPattern;
	}

	public Stack<MethodIdentifier> getOrCreateFinalizerStackForThread(Integer threadId) {
		Stack<MethodIdentifier> stack = threadIdToFinalizerMethodStackMap.get(threadId);
		if (stack == null) {
			stack = new Stack<MethodIdentifier>();
			threadIdToFinalizerMethodStackMap.put(threadId, stack);
		}
		return stack;
	}
}
