package de.unisaarland.cs.st.tautoko.verify;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Random;
import java.util.Set;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.exec.CommandLine;
import org.apache.commons.exec.ExecuteException;
import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;
import org.apache.tools.ant.filters.StringInputStream;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

import de.unisaarland.cs.st.adabu.core.CallStack;
import de.unisaarland.cs.st.adabu.core.InvokedMethod;
import de.unisaarland.cs.st.adabu.core.deep.miner.TypeSetDeepTransitiveModelMiner;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocation;
import de.unisaarland.cs.st.adabu.trans.model.MethodInvocationSet;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectModel;
import de.unisaarland.cs.st.adabu.trans.model.TransitiveObjectState;
import de.unisaarland.cs.st.adabu.util.Pair;
import de.unisaarland.cs.st.adabu.util.asm.ASMTreeUtilities;
import de.unisaarland.cs.st.adabu.util.asm.ASMUtilities;
import de.unisaarland.cs.st.adabu.util.asm.MethodIdentifier;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphEdge;
import de.unisaarland.cs.st.adabu.util.datastructures.GraphNode;
import de.unisaarland.cs.st.adabu.util.io.ObjectReaderWriter;
import de.unisaarland.cs.st.adabu.util.source.ClassFile;
import de.unisaarland.cs.st.adabu.util.source.ClassFileSource;
import de.unisaarland.cs.st.adabu.util.source.ClassPathClassFileSource;
import de.unisaarland.cs.st.ibugs.util.Util;
import de.unisaarland.cs.st.ibugs.util.Util.ExecutionResult;
import de.unisaarland.cs.st.tautoko.completion.BasicModifierGenerator;
import de.unisaarland.cs.st.tautoko.completion.JavaClass;
import de.unisaarland.cs.st.tautoko.completion.ModelCompleter;
import de.unisaarland.cs.st.tautoko.instrument.IModifier;

public class VerifierTester {

	private static final int TIMEOUT_COMPILE = 50000;
	private File workingDirectory;
	private File tmpDirectory;
	private File resultDirectory;
	private String testCommand;
	private String appClassPath;
	private String compileClassPath;
	private Set<String> acceptedTypes;
	private int heapPathLength;
	private int timeout;
	private static final Logger logger = Logger.getLogger(VerifierTester.class);
	private Random random = new Random();
	private IVerifierDriver verifierDriver;
	private boolean useClassNamesForComplexFields = false;
	private String spec;
	private TransitiveObjectModel specificationModel;
	private boolean falsePositives;

	public VerifierTester(File workingDirectory, File tmpDirectory, File resultDirectory, String testCommand,
	                      String appClassPath, String compileClassPath, String acceptedType, 
	                      int heapPathLength, TransitiveObjectModel specificationModel, int timeout,
	                      IVerifierDriver verifierDriver, String spec, boolean useClassNamesForComplexFields,
	                      boolean falsePositives) {
		this.workingDirectory = workingDirectory;
		this.tmpDirectory = tmpDirectory;
		this.testCommand = testCommand;
		this.resultDirectory = resultDirectory;
		this.appClassPath = appClassPath;
		this.compileClassPath = compileClassPath;
		this.acceptedTypes = Collections.singleton(acceptedType);
		this.heapPathLength = heapPathLength;
		this.specificationModel = specificationModel;
		this.timeout = timeout;
		this.verifierDriver = verifierDriver;
		this.spec = spec;
		this.useClassNamesForComplexFields = useClassNamesForComplexFields;
		this.falsePositives = falsePositives;
	}

	public void test() throws IOException, ClassNotFoundException, InstantiationException, IllegalAccessException, SAXException {
		ClassFileSource source = new ClassPathClassFileSource(appClassPath);
		File traceFile = new File(tmpDirectory + "/trace.ser");
		logger.info("Tracing initial models.");
		ExecutionResult executionResult = ModelCompleter.traceTests(testCommand, traceFile, workingDirectory, timeout);
		String[] types = getTypesFromSpec(spec);
		if (ModelCompleter.isTestOk(executionResult)) {
			List<TransitiveObjectModel> initialModels = mineModels(traceFile);
			ModelCompleter.abstractModels(initialModels, source, useClassNamesForComplexFields);
			Map<String, Set<TransitiveObjectModel>> minedTypeToModelMap = ModelCompleter.getTypeToModelMap(initialModels);
			logger.info("Mined " + initialModels.size() + " initial models for " + minedTypeToModelMap.size() + " types.");
			List<ModificationInformation> modificationList = getModifications(source, minedTypeToModelMap, types);
			Set<ModificationInformation> modifications = new HashSet<ModificationInformation>(modificationList);
			logger.info("Got " + modifications.size() + " modifications in total.");
			processModifications(modifications, source);
		} else {
			logger.info("Initial models not okay. Process returned with " + executionResult.getResult() + ". Output follows.");
			logger.info(executionResult.getOut());
		}
	}

	private void processModifications(Collection<ModificationInformation> modifications, ClassFileSource source) throws IOException, ClassNotFoundException, InstantiationException, IllegalAccessException {
		int modificationCounter = 0;
		for (ModificationInformation modification: modifications) {
			modificationCounter++;
			/*if (modificationCounter != 34) {
				logger.fatal("Skipping modification " + modificationCounter);
				continue;
			}*/
			/*String content = modification.getModifier().generateClass().getContent();
			if (content.indexOf("IMAPServer") == -1 || content.indexOf("expunge") == -1 || content.indexOf("subscribe") == -1) {
				System.out.println("Meep");
				modificationCounter++;
				continue;
			}*/
			logger.info("Processing modification " + modificationCounter + " out of " + modifications.size());
			FileUtils.cleanDirectory(tmpDirectory);
			Set<String> instrumentedClasses = modification.getModifier().getInstrumentedClassNames();
			JavaClass javaClass = modification.getModifier().generateClass();
			File javaFile = writeJavaClass(javaClass, tmpDirectory);
			ExecutionResult result = compileClasses(Collections.singletonList(javaFile), tmpDirectory, compileClassPath);
			if (result.getResult() == 0) {
				logger.info("Successfully compiled classes.");
				File classFile = new File(javaFile.getPath().replace(".java", ".class"));
				try {
					IModifier modifier = loadModifier(classFile, javaClass.getPackageName().replace('/', '.') + "." + javaClass.getClassName());
					File modifiedClassesDirectory = new File(tmpDirectory.getPath() + "/modified");
					modifiedClassesDirectory.mkdirs();
					Map<String, byte[]> modifiedClasses = modifyClasses(modifier, source, instrumentedClasses);
					for (String className : modifiedClasses.keySet()) {
						writeClassFile(className, modifiedClasses.get(className), modifiedClassesDirectory);
					}
					logger.info("Running modified test to look for exceptions.");
					ExecutionResult modifiedTestResult = runModifiedTest(modifiedClassesDirectory);
					if (!(modifiedTestResult.getResult() == Util.RETURN_CODE_TIMEOUT)) {
						File modifierFile = new File(resultDirectory + "/" + modificationCounter+ ".modifier");
						FileUtils.writeStringToFile(modifierFile, modification.getModifier().generateClass().getContent());
						File testOutputFile = new File(resultDirectory + "/" + modificationCounter + ".testout");
						FileUtils.writeStringToFile(testOutputFile, modifiedTestResult.getOut().toString());	
						List<List<Statement>> stackTraces = getStackTracesFromOutput(modifiedTestResult.getOut());
						if (stackTraces.size() > 0) {
							if (falsePositives) {
								logger.info("Trace raises an exception and will be ignored.");
								continue;
							} else {
								logger.info("Trace raises an exception and will be processed.");
							}
							logger.info("Modification: " + modification);
							logger.info("Test output: " + modifiedTestResult.getOut());
							List<Pair<MethodInvocation, Integer>> invocationsThatTriggerExceptions = new Vector<Pair<MethodInvocation,Integer>>();
							for (MethodInvocation invocation : modification.getInvocations()) {
								InvokedMethod callingMethod = invocation.getCallStack().getCallingMethod();
								logger.info("Checking invocation " + callingMethod);
								int lineNumber = getLineNumberForInvocation(invocation, source);
								List<Statement> matchingStackTraceEntries = getMatchingStackTraceEntries(stackTraces, callingMethod.getIdentifier(), lineNumber);
								if (matchingStackTraceEntries.size() > 0) {
									invocationsThatTriggerExceptions.add(new Pair<MethodInvocation, Integer>(invocation, lineNumber));
								} 
							}
							if (invocationsThatTriggerExceptions.size() > 0) {
								logger.info("Found " + invocationsThatTriggerExceptions.size() + " invocations that trigger exceptions.");
								testVerifierForBug(modification, modifiedClassesDirectory, invocationsThatTriggerExceptions, modifiedTestResult, modificationCounter);
							} else {
								List<Statement> firstInvocationStatements = getFirstInvocationOfMethodInType(stackTraces, modification.getType(), source);
								if (firstInvocationStatements.size() > 0) {
									logger.info("Found " + firstInvocationStatements.size() + " invocations in stack trace:");
									for (Statement statement : firstInvocationStatements) {
										logger.info(statement);
									}
									testVerifierForStackTraceBug(modification, modifiedClassesDirectory, firstInvocationStatements, modifiedTestResult, modificationCounter);
								} else {
									logger.info("No matching stack trace entries found at all.");
								}
							}	
						} else {
							if (falsePositives) {
								logger.info("Trace does not contain an exception. Will look for false positives.");
								List<Statement> firstInvocationStatements = new Vector<Statement>();
								for (String className : modification.modifier.getInstrumentedClassNames()) {
									firstInvocationStatements.add(new Statement(className, "fakeMethod", -1));
								}
								testVerifierForStackTraceBug(modification, modifiedClassesDirectory, firstInvocationStatements, modifiedTestResult, modificationCounter);
							} else {
								logger.info("No exception found. This modification is skipped.");
							}
						}
					} else {
						logger.info("Test ran into timeout. Will skip this modification.");
					}
				} catch (Exception processingException) {
					logger.error("An error occured processing modification " + modificationCounter);
				}
			} else {
				logger.error("Unable to compile classes. Output follows:");
				logger.error(result.getOut());
			}
		}
		logger.info("Done processing modifications.");
	}

	private Set<String> getSuperTypes(String type, ClassFileSource source) throws IOException {
		Set<String> result = new HashSet<String>();
		result.add(type);
		do {
			try {
				ClassFile classFile = source.getClassForName(type);
				type = ASMUtilities.getSuperClassName(classFile.getClassBytes());
				if (type != null) {
					result.add(type);
				}
			} catch (NoSuchElementException ex) {
				logger.warn("Unable to resolve class " + type);
				return result;
			}
		} while (type != null);
		return result;
	}

	private List<Statement> getFirstInvocationOfMethodInType(List<List<Statement>> stackTraces, String type, ClassFileSource source) throws IOException {
		Set<Statement> result = new HashSet<Statement>();
		Set<String> types = getSuperTypes(type, source);
		for (List<Statement> stackTrace : stackTraces) {
			int firstCallIndex = -1;
			for (int stackTraceIndex = stackTrace.size() - 1; stackTraceIndex > 0; stackTraceIndex--) {
				Statement currentStatement = stackTrace.get(stackTraceIndex);
				if (types.contains(currentStatement.className)) {
					firstCallIndex = stackTraceIndex;
					break;
				}
			}
			if (firstCallIndex != -1) {
				if (firstCallIndex < stackTrace.size() - 1) {
					Statement callingStatement = stackTrace.get(firstCallIndex + 1);
					result.add(callingStatement);
				}
			}
		}
		return new Vector<Statement>(result);
	}

	private void testVerifierForStackTraceBug(ModificationInformation modification, File modifiedClassFileDirectory, 
	                                          List<Statement> statements, ExecutionResult testOutput, 
	                                          int modificationNumber) throws IOException {
		String classPath = modifiedClassFileDirectory.getAbsolutePath() + ":" + appClassPath;
		Pair<ExecutionResult, List<Violation>> verifierResult = verifierDriver.testVerifierOnMethods(modification.getModifier().getInstrumentedMethods(), classPath, spec, tmpDirectory);
		List<Violation> violations = verifierResult.getV();
		List<Violation> foundViolations = getViolationsThatCorrespondToStackTrace(violations, statements);
		logger.info(foundViolations.size() + " out of " + violations.size() + " reported violations correspond to stack trace entries.");
		saveResult(modification, testOutput, modificationNumber, verifierResult, foundViolations);
	}

	private List<Violation> getViolationsThatCorrespondToStackTrace(List<Violation> violations, List<Statement> statements) {
		List<Violation> result = new Vector<Violation>();
		for (Violation violation : violations) {
			for (Statement statement : statements) {
				if (violation.getClassName().equals(statement.className) && violation.getLineNumber() == statement.lineNumber) {
					result.add(violation);
				}
			}
		}
		return result;
	}

	private void testVerifierForBug(ModificationInformation modification, File modifiedClassFileDirectory, 
	                                List<Pair<MethodInvocation, Integer>> invocations, ExecutionResult testOutput, 
	                                int modificationNumber) throws IOException {
		String classPath = modifiedClassFileDirectory.getAbsolutePath() + ":" + appClassPath;
		Pair<ExecutionResult, List<Violation>> verifierResult = verifierDriver.testVerifierOnMethods(modification.modifier.getInstrumentedMethods(), classPath, spec, tmpDirectory);
		List<Violation> violations = verifierResult.getV();
		Map<Violation, Pair<MethodInvocation, Integer>> violationsThatCorrespondToModifications = getViolationsThatCorrespondToModifications(violations, invocations);
		logger.info(violationsThatCorrespondToModifications.size() + " out of " + violations.size() + " reported violations correspond to modifications.");
		saveResult(modification, testOutput, modificationNumber, verifierResult, new Vector<Violation>(violationsThatCorrespondToModifications.keySet()));
	}

	private void saveResult(ModificationInformation information, ExecutionResult testOutput, int modificationNumber,
	                        Pair<ExecutionResult, List<Violation>> verifierResult, List<Violation> foundViolations) throws IOException {
		saveCommonResults(information, testOutput, modificationNumber, verifierResult);
		File violationsFoundFile = new File(resultDirectory + "/" + modificationNumber + ".foundviolations");
		ObjectReaderWriter<List<Violation>> violationWriter = new ObjectReaderWriter<List<Violation>>();
		violationWriter.write(foundViolations, violationsFoundFile);
	}

	private void saveCommonResults(ModificationInformation information,
	                               ExecutionResult testOutput, int modificationNumber,
	                               Pair<ExecutionResult, List<Violation>> verifierResult)
	throws IOException {
		File verifierOutputFile = new File(resultDirectory + "/" + modificationNumber + ".verifierout");
		FileUtils.writeStringToFile(verifierOutputFile, verifierResult.getU().toString());
		ObjectReaderWriter<List<Violation>> violationWriter = new ObjectReaderWriter<List<Violation>>();
		File violationOutputFile = new File(resultDirectory + "/" + modificationNumber + ".allviolations");
		violationWriter.write(verifierResult.getV(), violationOutputFile);
	}

	private Map<Violation, Pair<MethodInvocation, Integer>> getViolationsThatCorrespondToModifications(List<Violation> violations, List<Pair<MethodInvocation, Integer>> modifications) {
		Map<Violation, Pair<MethodInvocation, Integer>> result = new HashMap<Violation, Pair<MethodInvocation,Integer>>();
		for (Violation violation: violations) {
			if (violation.getLineNumber() != -1) {
				for(Pair<MethodInvocation, Integer> modification: modifications) {
					if (violation.getLineNumber() == modification.getV() && 
							violation.getClassName().equals(modification.getU().getCallStack().getCallingMethod().getIdentifier().getClassName())) {
						result.put(violation, modification);
						break;
					}
				}
			}
		}
		return result;
	}

	private List<Statement> getMatchingStackTraceEntries(List<List<Statement>> stackTraces, MethodIdentifier identifier, int lineNumber) {
		List<Statement> result = new Vector<Statement>();
		for (List<Statement> stackTrace : stackTraces) {
			for (Statement statement : stackTrace) {
				String statementClassName = statement.className.replace('.', '/');
				if (statementClassName.equals(identifier.getClassName()) && identifier.getMethodName().equals(statement.methodName)) {
					if (statement.lineNumber == -1) {
						//we don't have line number information in the stack trace, assume that it's okay to include the call
						result.add(statement);
					} else if (statement.lineNumber == lineNumber) {
						result.add(statement);
					}
				}
			}
		}
		return result;
	}

	public static int getLineNumberForInvocation(MethodInvocation invocation, ClassFileSource source) throws NoSuchElementException, IOException {
		CallStack callStack = invocation.getCallStack();
		InvokedMethod callingMethod = callStack.getCallStack().get(callStack.size() - 2);
		ClassFile classFile = source.getClassForName(callingMethod.getIdentifier().getClassName());
		ClassNode classNode = ASMTreeUtilities.getClassNode(classFile.getClassBytes());
		Map<MethodIdentifier, InsnList> methodByteCode = ASMTreeUtilities.getInstructionsForMethods(classNode);
		InsnList methodInstructions = methodByteCode.get(callingMethod.getIdentifier());
		Map<AbstractInsnNode, Integer> lineNumberMapping = ASMTreeUtilities.getLineNumberMapping(methodInstructions);
		MethodInsnNode methodInsnNode = ASMTreeUtilities.getCallSiteInstruction(methodInstructions, callStack.getActiveMethod().getCallSiteId());
		logger.info("Mapping invocation of " + callStack.getActiveMethod().getIdentifier() + " to call to " + methodInsnNode.owner + "."  + methodInsnNode.name + methodInsnNode.desc);
		return lineNumberMapping.get(methodInsnNode);
	}

	private static Pattern withDebugInformationPattern = Pattern.compile(".*at ([^\\(]*)\\(([^:]*):([^\\)]*)\\)", Pattern.DOTALL);

	private static Pattern withoutDebugInformationPattern = Pattern.compile(".*at ([^\\(]*)\\(.*", Pattern.DOTALL);

	public static List<List<Statement>> getStackTracesFromOutput(String output) {
		List<List<Statement>> stackTraces = new Vector<List<Statement>>();
		List<Statement> currentStackTrace = null;
		String[] lines = output.split("\n");
		boolean lastLineStackTrace = false;
		for (String line : lines) {
			String methodCall = null;
			int lineNumber = -1;
			Matcher matcher = withDebugInformationPattern.matcher(line);
			if (matcher.matches()) {
				methodCall = matcher.group(1);
				lineNumber = Integer.parseInt(matcher.group(3));
			} else {
				matcher = withoutDebugInformationPattern.matcher(line);
				if (matcher.matches()) {
					methodCall = matcher.group(1);
				}
			}
			if (methodCall != null) {
				if (!lastLineStackTrace) {
					lastLineStackTrace = true;
					currentStackTrace = new Vector<Statement>();
					stackTraces.add(currentStackTrace);
				}
				int lastDotIndex = methodCall.lastIndexOf('.');
				if (lastDotIndex != -1) {
					String className = methodCall.substring(0, lastDotIndex).replace('.', '/');
					String methodName = methodCall.substring(lastDotIndex + 1);
					Statement statement = new Statement(className, methodName, lineNumber);
					currentStackTrace.add(statement);
				}
			} else {
				lastLineStackTrace = false;
			}
		}
		return stackTraces;
	}

	public static void writeClassFile(String fullyQualifiedClassName, byte[] bytes, File baseDirectory) throws IOException {
		String packageName = "";
		String className = fullyQualifiedClassName;
		int lastIndex = fullyQualifiedClassName.indexOf('/');
		if (lastIndex != -1) {
			packageName = fullyQualifiedClassName.substring(0, lastIndex);
			className = fullyQualifiedClassName.substring(lastIndex + 1);
		}
		File classFileDirectory = new File(baseDirectory + "/" + packageName);
		classFileDirectory.mkdirs();
		File classFile = new File(classFileDirectory + "/" + className + ".class");
		FileUtils.writeByteArrayToFile(classFile, bytes);
	}

	public Map<String, byte[]> modifyClasses(IModifier modifier, ClassFileSource source, Collection<String> instrumentedClasses) throws NoSuchElementException, IOException {
		Map<String, byte[]> modifiedClasses = new HashMap<String, byte[]>();
		for (String className : instrumentedClasses) {
			ClassFile classFile = source.getClassForName(className);
			byte[] originalClassBytes = classFile.getClassBytes();
			byte[] transformedBytes = modifier.transform(className, originalClassBytes);
			modifiedClasses.put(className, transformedBytes);
		}
		return modifiedClasses;
	}

	public IModifier loadModifier(File modifierFile, final String className) throws IOException, ClassNotFoundException, InstantiationException, IllegalAccessException {
		final byte[] classBytes = FileUtils.readFileToByteArray(modifierFile);
		ClassLoader myClassLoader = new ClassLoader(getClass().getClassLoader()) {
			@Override
			public Class<?> loadClass(String name) throws ClassNotFoundException {
				if (name.equals(className)) {
					return defineClass(name, classBytes, 0, classBytes.length);
				} else {
					return getParent().loadClass(name);
				}
			}
		};
		Class modifierClass = myClassLoader.loadClass(className);
		return (IModifier) modifierClass.newInstance();
	}

	public static File writeJavaClass(JavaClass javaClass, File tmpDirectory) throws IOException {
		File packageDirectory = new File(tmpDirectory.getAbsolutePath() + File.separator + javaClass.getPackageName().replace('.', '/'));
		packageDirectory.mkdirs();
		File classFile = new File(packageDirectory.getAbsolutePath() + File.separator + javaClass.getClassName() + ".java");
		FileUtils.writeStringToFile(classFile, javaClass.getContent(), null);
		return classFile;
	}

	public static ExecutionResult compileClasses(List<File> javaFiles, File tmpDirectory, String compileClassPath) throws IOException {
		String command;
		if (System.getProperty("os.name").equalsIgnoreCase("linux")) {
			command = "/usr/lib/jvm/jdk1.6.0/bin/javac";
		} else {
			command = "/scratch2/dallmeier/soylatte/bin/javac";
		}
		List<String> params = new Vector<String>();
		params.add("-cp");
		params.add(compileClassPath + ":.");
		for (File javaFile : javaFiles) {
			params.add(javaFile.getPath());
		}
		String[] paramsArray = params.toArray(new String[params.size()]);
		return Util.runCommand(command, paramsArray, tmpDirectory, logger, TIMEOUT_COMPILE);
	}

	private List<String> getInstrumentedClasses(Collection<BasicModifierGenerator> generators) {
		List<String> result = new Vector<String>();
		for (BasicModifierGenerator generator : generators) {
			result.addAll(generator.getInstrumentedClassNames());
		}
		return result;
	}

	private List<JavaClass> getJavaClasses(Collection<BasicModifierGenerator> generators) {
		List<JavaClass> result = new Vector<JavaClass>(generators.size());
		for (BasicModifierGenerator generator : generators) {
			result.add(generator.generateClass());
		}
		return result;
	}

	private boolean isAppropriateType(String[] specTypes, Collection<String> typeInterfaces) {
		for (String typeType: typeInterfaces) {
			for (String specType: typeInterfaces) {
				if (typeType.equals(specType)) {
					return true;
				}
			}
		}
		return false;
	}

	public static Collection<String> getInterfaces(String className, ClassFileSource source) {
		try {
			ClassFile classFile = source.getClassForName(className);
			Collection<String> result = new Vector<String> (Arrays.asList(ASMUtilities.getInterfaces(classFile.getClassBytes())));
			result.add(className);
			return result;
		} catch (Exception e) {
			logger.warn("Unable to determine interfaces for class " + className, e);
			Vector<String> result = new Vector<String>();
			result.add(className);
			return result;
		}
	}

	private List<ModificationInformation> getModifications(ClassFileSource source, Map<String, Set<TransitiveObjectModel>> minedTypeToModelMap, String[] types)
	throws IOException {
		List<ModificationInformation> modifications = new Vector<ModificationInformation>();
		for (String type : minedTypeToModelMap.keySet()) {
			Collection<String> typeInterfaces = getInterfaces(type, source); 
			logger.info("Type interfaces: " + typeInterfaces);
			logger.info("Specification types: " + Arrays.toString(types));
			if (isAppropriateType(types, typeInterfaces)) {
				logger.info("Processing type " + type + ".");
				Set<TransitiveObjectModel> models = minedTypeToModelMap.get(type);
				for (TransitiveObjectModel model : models) {
					Set<TransitiveObjectState> nonStartingStates = ModelCompleter.getAllNonStartingStates(model);
					Set<MethodIdentifier> allIdentifiers = ModelCompleter.getAllPublicNonStaticMethods(model.getClassName(), source);
					Map<MethodIdentifier, Set<TransitiveObjectState>> methodToInvokedStatesMap = ModelCompleter.getMethodToInvokedStatesMap(Collections.singletonList(model));
					for (MethodIdentifier currentMethod : allIdentifiers) {
						if (!currentMethod.isConstructor()) {
							//first look for direct matches
							/*Set<TransitiveObjectState> statesToInvestigate = getErrorStatesForMethod(currentMethod, specificationModel, false);
							if (statesToInvestigate.isEmpty()) {
								logger.info("Found no direct error state matches. Will look for matches without class name.");
								statesToInvestigate = getErrorStatesForMethod(currentMethod, specificationModel, true);
							}*/
							Set<TransitiveObjectState> statesToInvestigate = new HashSet<TransitiveObjectState>(nonStartingStates);
							Set<TransitiveObjectState> originatingStates = ModelCompleter.getAllOriginatingStates(model, currentMethod);
							Set<TransitiveObjectState> exercisedStates = methodToInvokedStatesMap.get(currentMethod);
							if (exercisedStates == null) {
								exercisedStates = new HashSet<TransitiveObjectState>();
								methodToInvokedStatesMap.put(currentMethod, exercisedStates);
							}
							statesToInvestigate.removeAll(originatingStates);
							statesToInvestigate.removeAll(exercisedStates);
							if (!statesToInvestigate.isEmpty()) {
								logger.info("Found " + statesToInvestigate.size() + " states to investigate for method " + currentMethod);
								for (TransitiveObjectState stateToInvestigate : statesToInvestigate) {
									try {
										if (model.getModel().containsNodeForData(stateToInvestigate)) {
											if (!currentMethod.hasParameters()) {
												List<Pair<BasicModifierGenerator, MethodInvocation>> insertMethodCallJavaClasses = ModelCompleter.insertMethodCall(model, currentMethod, stateToInvestigate, source);
												for (Pair<BasicModifierGenerator, MethodInvocation> pair : insertMethodCallJavaClasses) {
													ModificationInformation modificationInformation = new ModificationInformation(Collections.singleton(pair.getV()), pair.getU(), type);
													modificationInformation.setAdd(true);
													modifications.add(modificationInformation);
												}	
											} else {
												for (GraphNode<TransitiveObjectState, MethodInvocationSet> node: model.getModel().getNodes()) {
													List<GraphEdge<TransitiveObjectState, MethodInvocationSet>> edges = node.getOutgoingEdges();
													for (GraphEdge<TransitiveObjectState, MethodInvocationSet> graphEdge : edges) {
														MethodInvocationSet methodInvocationSet = graphEdge.getData();
														Set<MethodInvocation> allInvocations = methodInvocationSet.getAllInvocationsOfMethod(currentMethod);
														if (allInvocations.size() > 0) {
															List<BasicModifierGenerator> supressMethodCallsJavaClasses = ModelCompleter.supressMethodCalls(model, currentMethod, stateToInvestigate, source, Collections.singleton(node.getData()));
															if (supressMethodCallsJavaClasses != null && supressMethodCallsJavaClasses.size() > 0) {
																ModificationInformation information = new ModificationInformation(allInvocations, supressMethodCallsJavaClasses.get(0), type);
																modifications.add(information);
															}
														}
													}
												}	
											}
										}
									} catch (Exception processingException) {
										logger.warn("Unable to create modification for state " + stateToInvestigate, processingException);
									}
								}
							} else {
								logger.info("Found no error states for model.");
							}
						}
					}
				}
			} else {
				logger.info("Skipping type " + type + " because there is no specification available.");
			}
		}
		return modifications;
	}

	public static class ModificationInformation {
		private Set<MethodInvocation> invocations;
		private boolean isAdd;
		private String type;
		private BasicModifierGenerator modifier;
		private JavaClass javaClass;
		public boolean isAdd() {
			return isAdd;
		}
		public void setAdd(boolean isAdd) {
			this.isAdd = isAdd;
		}
		public ModificationInformation(Set<MethodInvocation> invocations, BasicModifierGenerator modifier, String type) {
			super();
			this.invocations = invocations;
			this.modifier = modifier;
			this.type = type;
			this.javaClass = modifier.generateClass();
		}
		public String getType()	 {
			return type;
		}
		public Set<MethodInvocation> getInvocations() {
			return invocations;
		}
		public BasicModifierGenerator getModifier() {
			return modifier;
		}
		@Override
		public boolean equals(Object o) {
			if (o != null && o instanceof ModificationInformation) {
				return ((ModificationInformation) o).javaClass.equals(this.javaClass);
			} else {
				return false;
			}
		}
		@Override
		public int hashCode() {
			return javaClass.hashCode();
		}
		public String toString() {
			return javaClass.getContent();
		}
	}

	public static MethodInvocation getMethodInvocation(List<GraphEdge<TransitiveObjectState, MethodInvocationSet>> edges, MethodIdentifier identifier) {
		for (GraphEdge<TransitiveObjectState, MethodInvocationSet> graphEdge : edges) {
			MethodInvocationSet invocationSet = graphEdge.getData();
			Set<MethodInvocation> invocations = invocationSet.getAllInvocationsOfMethod(identifier);
			if (invocations.size() > 0) {
				return invocations.iterator().next();
			}
		}
		return null;
	}

	public ExecutionResult runModifiedTest(File modifiedClassesDirectory) throws ExecuteException, IOException {
		String killIdentifier = Long.toString(random.nextLong());
		Vector<String> args = new Vector<String>();
		args.add("-cp");
		args.add(modifiedClassesDirectory.getAbsolutePath() + ":" + appClassPath);
		args.add("-Dkill.identifier=" + killIdentifier);
		return Util.runCommand(testCommand, args.toArray(new String[args.size()]), workingDirectory, logger, timeout, killIdentifier);
	}

	public static Set<MethodInvocation> getInvocationsWithMatchingNameAndDesc(MethodInvocationSet invocations, MethodIdentifier identifier) {
		Set<MethodInvocation> result = new HashSet<MethodInvocation>();
		for (MethodInvocation methodInvocation : invocations) {
			MethodIdentifier invokedIdentifier = methodInvocation.getIdentifier();
			if (invokedIdentifier.getMethodName().equals(identifier.getMethodName()) && invokedIdentifier.getSignature().equals(identifier.getSignature())) {
				result.add(methodInvocation);
			}
		}
		return result;
	}

	/*public static Set<TransitiveObjectState> getErrorStatesForMethod(MethodIdentifier identifier, TransitiveObjectModel model, boolean checkOnlyNameAndSignature) {
		Set<TransitiveObjectState> states = new HashSet<TransitiveObjectState>();
		for (GraphNode<TransitiveObjectState, MethodInvocationSet> node: model.getModel().getNodes()) {
			boolean hasOnlyExceptionalInvocations = true;
			boolean hasInvocations = false;
			for (GraphEdge<TransitiveObjectState, MethodInvocationSet> edge : node.getOutgoingEdges()) {
				Set<MethodInvocation> methodInvocations;
				if (checkOnlyNameAndSignature) {
					methodInvocations = getInvocationsWithMatchingNameAndDesc(edge.getData(), identifier);
				} else {
					methodInvocations = edge.getData().getAllInvocationsOfMethod(identifier);
				}
				if (methodInvocations.size() > 0) {
					hasInvocations = true;
					for (MethodInvocation methodInvocation : methodInvocations) {
						if (!methodInvocation.exceptionRaised()) {
							hasOnlyExceptionalInvocations = false;
							break;
						}
					}

				}
			}
			if (hasInvocations && hasOnlyExceptionalInvocations) {
				states.add(node.getData());
			}	
		}
		return states;
	}*/

	private Map<String, String> parseTypesFromSpecs(List<String> specs) throws SAXException, IOException {
		Map<String, String> result = new HashMap<String, String>();
		for (String spec: specs) {
			String[] types = getTypesFromSpec(spec);
			for (String type : types) {
				result.put(type, spec);
			}
		}
		return result;
	}

	private String[] getTypesFromSpec(String spec) throws SAXException, IOException {
		XMLReader parser = XMLReaderFactory.createXMLReader();
		final List<String> types = new Vector<String>(3);
		parser.setContentHandler(new DefaultHandler() {
			@Override
			public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
				if (localName.equals("graph")) {
					String id = attributes.getValue("id");
					String[] typeArray = id.split(" ");
					for (String type : typeArray) {
						types.add(type);
					}
				}
			}
		});
		InputSource inputSource = new InputSource(new StringInputStream(spec));
		parser.parse(inputSource);
		return types.toArray(new String[types.size()]);
	}

	public List<TransitiveObjectModel> mineModels(File traceFile) throws IOException {
		TypeSetDeepTransitiveModelMiner miner = new TypeSetDeepTransitiveModelMiner(traceFile, heapPathLength, null, acceptedTypes);
		miner.readTrace(traceFile);
		List<TransitiveObjectModel> models = miner.getModels();
		return models;
	}

	private static class Statement {
		private String className;
		private String methodName;
		private int lineNumber;
		public Statement(String className, String methodName, int lineNumber) {
			super();
			this.className = className;
			this.methodName = methodName;
			this.lineNumber = lineNumber;
		}
		public String getClassName() {
			return className;
		}
		public String getMethodName() {
			return methodName;
		}
		public int getLineNumber() {
			return lineNumber;
		}

		public String toString() {
			return className + "." + methodName + "@" + lineNumber;
		}
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((className == null) ? 0 : className.hashCode());
			result = prime * result + lineNumber;
			result = prime * result + ((methodName == null) ? 0 : methodName.hashCode());
			return result;
		}
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Statement other = (Statement) obj;
			if (className == null) {
				if (other.className != null)
					return false;
			} else if (!className.equals(other.className))
				return false;
			if (lineNumber != other.lineNumber)
				return false;
			if (methodName == null) {
				if (other.methodName != null)
					return false;
			} else if (!methodName.equals(other.methodName))
				return false;
			return true;
		}
	}

}
